@@ -52,18 +52,16 @@
 #include "qcocoafiledialoghelper.h"
 #include "qcocoatheme.h"
 #include "qcocoainputcontext.h"
-#include "qmacmime.h"
+#include "qcocoamimetypes.h"
 #include "qcocoaaccessibility.h"
 
 #include <qpa/qplatformaccessibility.h>
 #include <QtCore/qcoreapplication.h>
 
-#include <QtPlatformSupport/private/qcoretextfontdatabase_p.h>
 #include <IOKit/graphics/IOGraphicsLib.h>
 
 static void initResources()
 {
-    Q_INIT_RESOURCE_EXTERN(qcocoaresources)
     Q_INIT_RESOURCE(qcocoaresources);
 }
 
@@ -83,12 +81,16 @@ QCocoaScreen::~QCocoaScreen()
 
 NSScreen *QCocoaScreen::osScreen() const
 {
-    return [[NSScreen screens] objectAtIndex:m_screenIndex];
+    NSArray *screens = [NSScreen screens];
+    return ((NSUInteger)m_screenIndex < [screens count]) ? [screens objectAtIndex:m_screenIndex] : nil;
 }
 
 void QCocoaScreen::updateGeometry()
 {
     NSScreen *nsScreen = osScreen();
+    if (!nsScreen)
+        return;
+
     NSRect frameRect = [nsScreen frame];
 
     if (m_screenIndex == 0) {
@@ -145,7 +147,8 @@ qreal QCocoaScreen::devicePixelRatio() const
 {
 #if MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_7
     if (QSysInfo::MacintoshVersion >= QSysInfo::MV_10_7) {
-        return qreal([osScreen() backingScaleFactor]);
+        NSScreen * screen = osScreen();
+        return qreal(screen ? [screen backingScaleFactor] : 1.0);
     } else
 #endif
     {
@@ -153,6 +156,46 @@ qreal QCocoaScreen::devicePixelRatio() const
     }
 }
 
+QWindow *QCocoaScreen::topLevelAt(const QPoint &point) const
+{
+    NSPoint screenPoint = qt_mac_flipPoint(point);
+
+    // Search (hit test) for the top-level window. [NSWidow windowNumberAtPoint:
+    // belowWindowWithWindowNumber] may return windows that are not interesting
+    // to Qt. The search iterates until a suitable window or no window is found.
+    NSInteger topWindowNumber = 0;
+    QWindow *window = 0;
+    do {
+        // Get the top-most window, below any previously rejected window.
+        topWindowNumber = [NSWindow windowNumberAtPoint:screenPoint
+                                    belowWindowWithWindowNumber:topWindowNumber];
+
+        // Continue the search if the window does not belong to this process.
+        NSWindow *nsWindow = [NSApp windowWithWindowNumber:topWindowNumber];
+        if (nsWindow == 0)
+            continue;
+
+        // Continue the search if the window does not belong to Qt.
+        if (![nsWindow conformsToProtocol:@protocol(QNSWindowProtocol)])
+            continue;
+
+        id<QNSWindowProtocol> proto = static_cast<id<QNSWindowProtocol> >(nsWindow);
+        QCocoaWindow *cocoaWindow = proto.helper.platformWindow;
+        if (!cocoaWindow)
+            continue;
+        window = cocoaWindow->window();
+
+        // Continue the search if the window is not a top-level window.
+        if (!window->isTopLevel())
+             continue;
+
+        // Stop searching. The current window is the correct window.
+        break;
+    } while (topWindowNumber > 0);
+
+    return window;
+}
+
 extern CGContextRef qt_mac_cg_context(const QPaintDevice *pdev);
 
 QPixmap QCocoaScreen::grabWindow(WId window, int x, int y, int width, int height) const
@@ -199,6 +242,7 @@ QPixmap QCocoaScreen::grabWindow(WId window, int x, int y, int width, int height
         int w = (width < 0 ? bounds.size.width : width) * devicePixelRatio();
         int h = (height < 0 ? bounds.size.height : height) * devicePixelRatio();
         QRect displayRect = QRect(x, y, w, h);
+        displayRect = displayRect.translated(qRound(-bounds.origin.x), qRound(-bounds.origin.y));
         QCFType<CGImageRef> image = CGDisplayCreateImageForRect(displays[i],
             CGRectMake(displayRect.x(), displayRect.y(), displayRect.width(), displayRect.height()));
         QPixmap pix(w, h);
@@ -209,14 +253,15 @@ QPixmap QCocoaScreen::grabWindow(WId window, int x, int y, int width, int height
         CGContextRelease(ctx);
 
         QPainter painter(&windowPixmap);
-        painter.drawPixmap(bounds.origin.x, bounds.origin.y, pix);
+        painter.drawPixmap(0, 0, pix);
     }
     return windowPixmap;
 }
 
+QCocoaIntegration *QCocoaIntegration::mInstance = 0;
+
 QCocoaIntegration::QCocoaIntegration()
     : mFontDb(new QCoreTextFontDatabase())
-    , mEventDispatcher(new QCocoaEventDispatcher())
     , mInputContext(new QCocoaInputContext)
 #ifndef QT_NO_ACCESSIBILITY
     , mAccessibility(new QCocoaAccessibility)
@@ -227,12 +272,16 @@ QCocoaIntegration::QCocoaIntegration()
     , mServices(new QCocoaServices)
     , mKeyboardMapper(new QCocoaKeyMapper)
 {
+    if (mInstance != 0)
+        qWarning("Creating multiple Cocoa platform integrations is not supported");
+    mInstance = this;
+
     initResources();
     QCocoaAutoReleasePool pool;
 
     qApp->setAttribute(Qt::AA_DontUseNativeMenuBar, false);
 
-    NSApplication *cocoaApplication = [QT_MANGLE_NAMESPACE(QNSApplication) sharedApplication];
+    NSApplication *cocoaApplication = [QNSApplication sharedApplication];
     qt_redirectNSApplicationSendEvent();
 
     if (qEnvironmentVariableIsEmpty("QT_MAC_DISABLE_FOREGROUND_APPLICATION_TRANSFORM")) {
@@ -256,12 +305,12 @@ QCocoaIntegration::QCocoaIntegration()
     if (!QCoreApplication::testAttribute(Qt::AA_MacPluginApplication)) {
 
         // Set app delegate, link to the current delegate (if any)
-        QT_MANGLE_NAMESPACE(QCocoaApplicationDelegate) *newDelegate = [QT_MANGLE_NAMESPACE(QCocoaApplicationDelegate) sharedDelegate];
+        QCocoaApplicationDelegate *newDelegate = [QCocoaApplicationDelegate sharedDelegate];
         [newDelegate setReflectionDelegate:[cocoaApplication delegate]];
         [cocoaApplication setDelegate:newDelegate];
 
         // Load the application menu. This menu contains Preferences, Hide, Quit.
-        QT_MANGLE_NAMESPACE(QCocoaMenuLoader) *qtMenuLoader = [[QT_MANGLE_NAMESPACE(QCocoaMenuLoader) alloc] init];
+        QCocoaMenuLoader *qtMenuLoader = [[QCocoaMenuLoader alloc] init];
         qt_mac_loadMenuNib(qtMenuLoader);
         [cocoaApplication setMenu:[qtMenuLoader menu]];
         [newDelegate setMenuLoader:qtMenuLoader];
@@ -269,17 +318,20 @@ QCocoaIntegration::QCocoaIntegration()
 
     updateScreens();
 
-    QMacPasteboardMime::initializeMimeTypes();
+    QMacInternalPasteboardMime::initializeMimeTypes();
+    QCocoaMimeTypes::initializeMimeTypes();
 }
 
 QCocoaIntegration::~QCocoaIntegration()
 {
+    mInstance = 0;
+
     qt_resetNSApplicationSendEvent();
 
     QCocoaAutoReleasePool pool;
     if (!QCoreApplication::testAttribute(Qt::AA_MacPluginApplication)) {
         // remove the apple event handlers installed by QCocoaApplicationDelegate
-        QT_MANGLE_NAMESPACE(QCocoaApplicationDelegate) *delegate = [QT_MANGLE_NAMESPACE(QCocoaApplicationDelegate) sharedDelegate];
+        QCocoaApplicationDelegate *delegate = [QCocoaApplicationDelegate sharedDelegate];
         [delegate removeAppleEventHandlers];
         // reset the application delegate
         [[NSApplication sharedApplication] setDelegate: 0];
@@ -289,12 +341,19 @@ QCocoaIntegration::~QCocoaIntegration()
     // Deleting the clipboard integration flushes promised pastes using
     // the mime converters - the ordering here is important.
     delete mCocoaClipboard;
-    QMacPasteboardMime::destroyMimeTypes();
+    QMacInternalPasteboardMime::destroyMimeTypes();
 
     // Delete screens in reverse order to avoid crash in case of multiple screens
     while (!mScreens.isEmpty()) {
         delete mScreens.takeLast();
     }
+
+    clearToolbars();
+}
+
+QCocoaIntegration *QCocoaIntegration::instance()
+{
+    return mInstance;
 }
 
 /*!
@@ -302,11 +361,17 @@ QCocoaIntegration::~QCocoaIntegration()
 */
 void QCocoaIntegration::updateScreens()
 {
-    NSArray *screens = [NSScreen screens];
+    NSArray *scrs = [NSScreen screens];
+    NSMutableArray *screens = [NSMutableArray arrayWithArray:scrs];
+    if ([screens count] == 0)
+        if ([NSScreen mainScreen])
+           [screens addObject:[NSScreen mainScreen]];
+    if ([screens count] == 0)
+        return;
     QSet<QCocoaScreen*> remainingScreens = QSet<QCocoaScreen*>::fromList(mScreens);
     QList<QPlatformScreen *> siblings;
     for (uint i = 0; i < [screens count]; i++) {
-        NSScreen* scr = [[NSScreen screens] objectAtIndex:i];
+        NSScreen* scr = [screens objectAtIndex:i];
         CGDirectDisplayID dpy = [[[scr deviceDescription] objectForKey:@"NSScreenNumber"] unsignedIntValue];
         // If this screen is a mirror and is not the primary one of the mirror set, ignore it.
         if (CGDisplayIsInMirrorSet(dpy)) {
@@ -343,6 +408,14 @@ void QCocoaIntegration::updateScreens()
         screen->setVirtualSiblings(siblings);
 }
 
+QCocoaScreen *QCocoaIntegration::screenAtIndex(int index)
+{
+    if (index >= mScreens.count())
+        updateScreens();
+
+    return mScreens.at(index);
+}
+
 bool QCocoaIntegration::hasCapability(QPlatformIntegration::Capability cap) const
 {
     switch (cap) {
@@ -353,6 +426,7 @@ bool QCocoaIntegration::hasCapability(QPlatformIntegration::Capability cap) cons
     case WindowMasks:
     case MultipleWindows:
     case ForeignWindows:
+    case RasterGLSurface:
         return true;
     default:
         return QPlatformIntegration::hasCapability(cap);
@@ -376,27 +450,27 @@ QPlatformBackingStore *QCocoaIntegration::createPlatformBackingStore(QWindow *wi
     return new QCocoaBackingStore(window);
 }
 
-QAbstractEventDispatcher *QCocoaIntegration::guiThreadEventDispatcher() const
+QAbstractEventDispatcher *QCocoaIntegration::createEventDispatcher() const
 {
-    return mEventDispatcher;
+    return new QCocoaEventDispatcher;
 }
 
-QPlatformFontDatabase *QCocoaIntegration::fontDatabase() const
+QCoreTextFontDatabase *QCocoaIntegration::fontDatabase() const
 {
     return mFontDb.data();
 }
 
-QPlatformNativeInterface *QCocoaIntegration::nativeInterface() const
+QCocoaNativeInterface *QCocoaIntegration::nativeInterface() const
 {
     return mNativeInterface.data();
 }
 
-QPlatformInputContext *QCocoaIntegration::inputContext() const
+QCocoaInputContext *QCocoaIntegration::inputContext() const
 {
     return mInputContext.data();
 }
 
-QPlatformAccessibility *QCocoaIntegration::accessibility() const
+QCocoaAccessibility *QCocoaIntegration::accessibility() const
 {
 #ifndef QT_NO_ACCESSIBILITY
     return mAccessibility.data();
@@ -405,12 +479,12 @@ QPlatformAccessibility *QCocoaIntegration::accessibility() const
 #endif
 }
 
-QPlatformClipboard *QCocoaIntegration::clipboard() const
+QCocoaClipboard *QCocoaIntegration::clipboard() const
 {
     return mCocoaClipboard;
 }
 
-QPlatformDrag *QCocoaIntegration::drag() const
+QCocoaDrag *QCocoaIntegration::drag() const
 {
     return mCocoaDrag.data();
 }
@@ -427,7 +501,7 @@ QPlatformTheme *QCocoaIntegration::createPlatformTheme(const QString &name) cons
     return QPlatformIntegration::createPlatformTheme(name);
 }
 
-QPlatformServices *QCocoaIntegration::services() const
+QCocoaServices *QCocoaIntegration::services() const
 {
     return mServices.data();
 }
@@ -442,9 +516,38 @@ QVariant QCocoaIntegration::styleHint(StyleHint hint) const
     return QPlatformIntegration::styleHint(hint);
 }
 
+Qt::KeyboardModifiers QCocoaIntegration::queryKeyboardModifiers() const
+{
+    return QCocoaKeyMapper::queryKeyboardModifiers();
+}
+
 QList<int> QCocoaIntegration::possibleKeys(const QKeyEvent *event) const
 {
     return mKeyboardMapper->possibleKeys(event);
 }
 
+void QCocoaIntegration::setToolbar(QWindow *window, NSToolbar *toolbar)
+{
+    if (NSToolbar *prevToolbar = mToolbars.value(window))
+        [prevToolbar release];
+
+    [toolbar retain];
+    mToolbars.insert(window, toolbar);
+}
+
+NSToolbar *QCocoaIntegration::toolbar(QWindow *window) const
+{
+    return mToolbars.value(window);
+}
+
+void QCocoaIntegration::clearToolbars()
+{
+    QHash<QWindow *, NSToolbar *>::const_iterator it = mToolbars.constBegin();
+    while (it != mToolbars.constEnd()) {
+        [it.value() release];
+        ++it;
+    }
+    mToolbars.clear();
+}
+
 QT_END_NAMESPACE
