@@ -65,6 +65,7 @@
 #include <QtCore/qmath.h>
 #include <QtCore/QThreadStorage>
 #include <QtCore/private/qsystemlibrary_p.h>
+#include <QtCore/private/qstringiterator_p.h>
 
 #include <QtCore/QDebug>
 
@@ -161,12 +162,32 @@ bool QWindowsFontEngine::hasCFFTable() const
     return GetFontData(hdc, MAKE_TAG('C', 'F', 'F', ' '), 0, 0, 0) != GDI_ERROR;
 }
 
+bool QWindowsFontEngine::hasCMapTable() const
+{
+    HDC hdc = m_fontEngineData->hdc;
+    SelectObject(hdc, hfont);
+    return GetFontData(hdc, MAKE_TAG('c', 'm', 'a', 'p'), 0, 0, 0) != GDI_ERROR;
+}
+
+bool QWindowsFontEngine::hasGlyfTable() const
+{
+    HDC hdc = m_fontEngineData->hdc;
+    SelectObject(hdc, hfont);
+    return GetFontData(hdc, MAKE_TAG('g', 'l', 'y', 'f'), 0, 0, 0) != GDI_ERROR;
+}
+
+bool QWindowsFontEngine::hasEbdtTable() const
+{
+    HDC hdc = m_fontEngineData->hdc;
+    SelectObject(hdc, hfont);
+    return GetFontData(hdc, MAKE_TAG('E', 'B', 'D', 'T'), 0, 0, 0) != GDI_ERROR;
+}
+
 void QWindowsFontEngine::getCMap()
 {
-    ttf = (bool)(tm.tmPitchAndFamily & TMPF_TRUETYPE);
+    ttf = (bool)(tm.tmPitchAndFamily & TMPF_TRUETYPE) || hasCMapTable();
 
-    // TMPF_TRUETYPE is not set for fonts with CFF tables
-    cffTable = !ttf && hasCFFTable();
+    cffTable = hasCFFTable();
 
     HDC hdc = m_fontEngineData->hdc;
     SelectObject(hdc, hfont);
@@ -199,80 +220,37 @@ void QWindowsFontEngine::getCMap()
     }
 }
 
-// ### Qt 5.1: replace with QStringIterator
-inline unsigned int getChar(const QChar *str, int &i, const int len)
+int QWindowsFontEngine::getGlyphIndexes(const QChar *str, int numChars, QGlyphLayout *glyphs) const
 {
-    uint uc = str[i].unicode();
-    if (QChar::isHighSurrogate(uc) && i < len-1) {
-        uint low = str[i+1].unicode();
-        if (QChar::isLowSurrogate(low)) {
-            uc = QChar::surrogateToUcs4(uc, low);
-            ++i;
-        }
-    }
-    return uc;
-}
-
-int QWindowsFontEngine::getGlyphIndexes(const QChar *str, int numChars, QGlyphLayout *glyphs, bool mirrored) const
-{
-    int i = 0;
     int glyph_pos = 0;
-    if (mirrored) {
+    {
 #if defined(Q_OS_WINCE)
         {
 #else
         if (symbol) {
-            for (; i < numChars; ++i, ++glyph_pos) {
-                unsigned int uc = getChar(str, i, numChars);
-                glyphs->glyphs[glyph_pos] = getTrueTypeGlyphIndex(cmap, uc);
-                if (!glyphs->glyphs[glyph_pos] && uc < 0x100)
-                    glyphs->glyphs[glyph_pos] = getTrueTypeGlyphIndex(cmap, uc + 0xf000);
-            }
-        } else if (ttf) {
-            for (; i < numChars; ++i, ++glyph_pos) {
-                unsigned int uc = getChar(str, i, numChars);
-                glyphs->glyphs[glyph_pos] = getTrueTypeGlyphIndex(cmap, QChar::mirroredChar(uc));
-            }
-        } else {
-#endif
-            wchar_t first = tm.tmFirstChar;
-            wchar_t last = tm.tmLastChar;
-
-            for (; i < numChars; ++i, ++glyph_pos) {
-                uint ucs = QChar::mirroredChar(getChar(str, i, numChars));
-                if (
-#ifdef Q_WS_WINCE
-                    tm.tmFirstChar > 60000 ||
-#endif
-                         ucs >= first && ucs <= last)
-                    glyphs->glyphs[glyph_pos] = ucs;
-                else
-                    glyphs->glyphs[glyph_pos] = 0;
-            }
-        }
-    } else {
-#if defined(Q_OS_WINCE)
-        {
-#else
-        if (symbol) {
-            for (; i < numChars; ++i, ++glyph_pos) {
-                unsigned int uc = getChar(str, i, numChars);
+            QStringIterator it(str, str + numChars);
+            while (it.hasNext()) {
+                const uint uc = it.next();
                 glyphs->glyphs[glyph_pos] = getTrueTypeGlyphIndex(cmap, uc);
                 if(!glyphs->glyphs[glyph_pos] && uc < 0x100)
                     glyphs->glyphs[glyph_pos] = getTrueTypeGlyphIndex(cmap, uc + 0xf000);
+                ++glyph_pos;
             }
         } else if (ttf) {
-            for (; i < numChars; ++i, ++glyph_pos) {
-                unsigned int uc = getChar(str, i, numChars);
+            QStringIterator it(str, str + numChars);
+            while (it.hasNext()) {
+                const uint uc = it.next();
                 glyphs->glyphs[glyph_pos] = getTrueTypeGlyphIndex(cmap, uc);
+                ++glyph_pos;
             }
         } else {
 #endif
             wchar_t first = tm.tmFirstChar;
             wchar_t last = tm.tmLastChar;
 
-            for (; i < numChars; ++i, ++glyph_pos) {
-                uint uc = getChar(str, i, numChars);
+            QStringIterator it(str, str + numChars);
+            while (it.hasNext()) {
+                const uint uc = it.next();
                 if (
 #ifdef Q_WS_WINCE
                     tm.tmFirstChar > 60000 ||
@@ -281,6 +259,7 @@ int QWindowsFontEngine::getGlyphIndexes(const QChar *str, int numChars, QGlyphLa
                     glyphs->glyphs[glyph_pos] = uc;
                 else
                     glyphs->glyphs[glyph_pos] = 0;
+                ++glyph_pos;
             }
         }
     }
@@ -299,7 +278,8 @@ int QWindowsFontEngine::getGlyphIndexes(const QChar *str, int numChars, QGlyphLa
 
 QWindowsFontEngine::QWindowsFontEngine(const QString &name,
                                HFONT _hfont, bool stockFontIn, LOGFONT lf,
-                               const QSharedPointer<QWindowsFontEngineData> &fontEngineData) :
+                               const QSharedPointer<QWindowsFontEngineData> &fontEngineData)
+    : QFontEngine(Win),
     m_fontEngineData(fontEngineData),
     _name(name),
     hfont(_hfont),
@@ -319,8 +299,7 @@ QWindowsFontEngine::QWindowsFontEngine(const QString &name,
     designAdvances(0),
     designAdvancesSize(0)
 {
-    if (QWindowsContext::verboseFonts)
-        qDebug("%s: font='%s', size=%ld", __FUNCTION__, qPrintable(name), lf.lfHeight);
+    qCDebug(lcQpaFonts) << __FUNCTION__ << name << lf.lfHeight;
     HDC hdc = m_fontEngineData->hdc;
     SelectObject(hdc, hfont);
     fontDef.pixelSize = -lf.lfHeight;
@@ -343,6 +322,8 @@ QWindowsFontEngine::QWindowsFontEngine(const QString &name,
     userData.insert(QStringLiteral("hFont"), QVariant::fromValue(hfont));
     userData.insert(QStringLiteral("trueType"), QVariant(bool(ttf)));
     setUserData(userData);
+
+    hasUnreliableOutline = hasGlyfTable() && hasEbdtTable();
 }
 
 QWindowsFontEngine::~QWindowsFontEngine()
@@ -360,9 +341,7 @@ QWindowsFontEngine::~QWindowsFontEngine()
         if (!DeleteObject(hfont))
             qErrnoWarning("%s: QFontEngineWin: failed to delete non-stock font... failed", __FUNCTION__);
     }
-    if (QWindowsContext::verboseFonts)
-        if (QWindowsContext::verboseFonts)
-            qDebug("%s: font='%s", __FUNCTION__, qPrintable(_name));
+    qCDebug(lcQpaFonts) << __FUNCTION__ << _name;
 
     if (!uniqueFamilyName.isEmpty()) {
         QPlatformFontDatabase *pfdb = QWindowsIntegration::instance()->fontDatabase();
@@ -370,23 +349,49 @@ QWindowsFontEngine::~QWindowsFontEngine()
     }
 }
 
+glyph_t QWindowsFontEngine::glyphIndex(uint ucs4) const
+{
+    glyph_t glyph;
+
+#if !defined(Q_OS_WINCE)
+    if (symbol) {
+        glyph = getTrueTypeGlyphIndex(cmap, ucs4);
+        if (glyph == 0 && ucs4 < 0x100)
+            glyph = getTrueTypeGlyphIndex(cmap, ucs4 + 0xf000);
+    } else if (ttf) {
+        glyph = getTrueTypeGlyphIndex(cmap, ucs4);
+#else
+    if (tm.tmFirstChar > 60000) {
+        glyph = ucs4;
+#endif
+    } else if (ucs4 >= tm.tmFirstChar && ucs4 <= tm.tmLastChar) {
+        glyph = ucs4;
+    } else {
+        glyph = 0;
+    }
+
+    return glyph;
+}
+
 HGDIOBJ QWindowsFontEngine::selectDesignFont() const
 {
     LOGFONT f = m_logfont;
     f.lfHeight = unitsPerEm;
+    f.lfWidth = 0;
     HFONT designFont = CreateFontIndirect(&f);
     return SelectObject(m_fontEngineData->hdc, designFont);
 }
 
 bool QWindowsFontEngine::stringToCMap(const QChar *str, int len, QGlyphLayout *glyphs, int *nglyphs, QFontEngine::ShaperFlags flags) const
 {
+    Q_ASSERT(glyphs->numGlyphs >= *nglyphs);
     if (*nglyphs < len) {
         *nglyphs = len;
         return false;
     }
 
     glyphs->numGlyphs = *nglyphs;
-    *nglyphs = getGlyphIndexes(str, len, glyphs, flags & RightToLeft);
+    *nglyphs = getGlyphIndexes(str, len, glyphs);
 
     if (!(flags & GlyphIndicesOnly))
         recalcAdvances(glyphs, flags);
@@ -427,8 +432,7 @@ void QWindowsFontEngine::recalcAdvances(QGlyphLayout *glyphs, QFontEngine::Shape
                 calculateTTFGlyphWidth(hdc, glyph, width);
                 designAdvances[glyph] = QFixed(width) / designToDevice;
             }
-            glyphs->advances_x[i] = designAdvances[glyph];
-            glyphs->advances_y[i] = 0;
+            glyphs->advances[i] = designAdvances[glyph];
         }
         if(oldFont)
             DeleteObject(SelectObject(hdc, oldFont));
@@ -436,8 +440,6 @@ void QWindowsFontEngine::recalcAdvances(QGlyphLayout *glyphs, QFontEngine::Shape
         for(int i = 0; i < glyphs->numGlyphs; i++) {
             unsigned int glyph = glyphs->glyphs[i];
 
-            glyphs->advances_y[i] = 0;
-
             if (glyph >= widthCacheSize) {
                 int newSize = (glyph + 256) >> 8 << 8;
                 widthCache = q_check_ptr((unsigned char *)realloc(widthCache,
@@ -445,9 +447,9 @@ void QWindowsFontEngine::recalcAdvances(QGlyphLayout *glyphs, QFontEngine::Shape
                 memset(widthCache + widthCacheSize, 0, newSize - widthCacheSize);
                 widthCacheSize = newSize;
             }
-            glyphs->advances_x[i] = widthCache[glyph];
+            glyphs->advances[i] = widthCache[glyph];
             // font-width cache failed
-            if (glyphs->advances_x[i] == 0) {
+            if (glyphs->advances[i].value() == 0) {
                 int width = 0;
                 if (!oldFont)
                     oldFont = SelectObject(hdc, hfont);
@@ -466,7 +468,7 @@ void QWindowsFontEngine::recalcAdvances(QGlyphLayout *glyphs, QFontEngine::Shape
                 } else {
                     calculateTTFGlyphWidth(hdc, glyph, width);
                 }
-                glyphs->advances_x[i] = width;
+                glyphs->advances[i] = width;
                 // if glyph's within cache range, store it for later
                 if (width > 0 && width < 0x100)
                     widthCache[glyph] = width;
@@ -676,6 +678,11 @@ void QWindowsFontEngine::getGlyphBearings(glyph_t glyph, qreal *leftBearing, qre
 }
 #endif // Q_CC_MINGW
 
+bool QWindowsFontEngine::hasUnreliableGlyphOutline() const
+{
+    return hasUnreliableOutline;
+}
+
 qreal QWindowsFontEngine::minLeftBearing() const
 {
     if (lbearing == SHRT_MIN)
@@ -778,46 +785,6 @@ qreal QWindowsFontEngine::minRightBearing() const
 #endif // Q_OS_WINCE
 }
 
-
-const char *QWindowsFontEngine::name() const
-{
-    return 0;
-}
-
-bool QWindowsFontEngine::canRender(const QChar *string,  int len)
-{
-    if (symbol) {
-        for (int i = 0; i < len; ++i) {
-            unsigned int uc = getChar(string, i, len);
-            if (getTrueTypeGlyphIndex(cmap, uc) == 0) {
-                if (uc < 0x100) {
-                    if (getTrueTypeGlyphIndex(cmap, uc + 0xf000) == 0)
-                        return false;
-                } else {
-                    return false;
-                }
-            }
-        }
-    } else if (ttf) {
-        for (int i = 0; i < len; ++i) {
-            unsigned int uc = getChar(string, i, len);
-            if (getTrueTypeGlyphIndex(cmap, uc) == 0)
-                return false;
-        }
-    } else {
-        while(len--) {
-            if (tm.tmFirstChar > string->unicode() || tm.tmLastChar < string->unicode())
-                return false;
-        }
-    }
-    return true;
-}
-
-QFontEngine::Type QWindowsFontEngine::type() const
-{
-    return QFontEngine::Win;
-}
-
 static inline double qt_fixed_to_double(const FIXED &p) {
     return ((p.value << 16) + p.fract) / 65536.0;
 }
@@ -1064,6 +1031,7 @@ bool QWindowsFontEngine::getSfntTableData(uint tag, uchar *buffer, uint *length)
     SelectObject(hdc, hfont);
     DWORD t = qbswap<quint32>(tag);
     *length = GetFontData(hdc, t, 0, buffer, *length);
+    Q_ASSERT(*length == GDI_ERROR || int(*length) > 0);
     return *length != GDI_ERROR;
 }
 
@@ -1085,7 +1053,7 @@ QWindowsNativeImage *QWindowsFontEngine::drawGDIGlyph(HFONT font, glyph_t glyph,
     int iw = gm.width.toInt();
     int ih = gm.height.toInt();
 
-    if (iw <= 0 || iw <= 0)
+    if (iw <= 0 || ih <= 0)
         return 0;
 
     bool has_transformation = t.type() > QTransform::TxTranslate;
@@ -1180,7 +1148,7 @@ glyph_metrics_t QWindowsFontEngine::alphaMapBoundingBox(glyph_t glyph, QFixed, c
 {
     int margin = 0;
     if (format == QFontEngine::Format_A32 || format == QFontEngine::Format_ARGB)
-        margin = glyphMargin(QFontEngineGlyphCache::Raster_RGBMask);
+        margin = glyphMargin(QFontEngine::Format_A32);
     glyph_metrics_t gm = boundingBox(glyph, matrix);
     gm.width += margin * 2;
     gm.height += margin * 2;
@@ -1253,7 +1221,7 @@ QImage QWindowsFontEngine::alphaRGBMapForGlyph(glyph_t glyph, QFixed, const QTra
     SystemParametersInfo(SPI_GETFONTSMOOTHINGCONTRAST, 0, &contrast, 0);
     SystemParametersInfo(SPI_SETFONTSMOOTHINGCONTRAST, 0, (void *) 1000, 0);
 
-    int margin = glyphMargin(QFontEngineGlyphCache::Raster_RGBMask);
+    int margin = glyphMargin(QFontEngine::Format_A32);
     QWindowsNativeImage *mask = drawGDIGlyph(font, glyph, margin, t, QImage::Format_RGB32);
     SystemParametersInfo(SPI_SETFONTSMOOTHINGCONTRAST, 0, (void *) quintptr(contrast), 0);
 
@@ -1286,15 +1254,11 @@ QFontEngine *QWindowsFontEngine::cloneWithSize(qreal pixelSize) const
     if (!uniqueFamilyName.isEmpty())
         request.family = uniqueFamilyName;
     request.pixelSize = pixelSize;
-    // Disable font merging, as otherwise createEngine will return a multi-engine
-    // instance instead of the specific engine we wish to clone.
-    request.styleStrategy |= QFont::NoFontMerging;
 
     QFontEngine *fontEngine =
-        QWindowsFontDatabase::createEngine(QChar::Script_Common, request, 0,
+        QWindowsFontDatabase::createEngine(request, 0,
                                            QWindowsContext::instance()->defaultDPI(),
-                                           false,
-                                           QStringList(), m_fontEngineData);
+                                           false, m_fontEngineData);
     if (fontEngine) {
         fontEngine->fontDef.family = actualFontName;
         if (!uniqueFamilyName.isEmpty()) {
@@ -1335,27 +1299,14 @@ void QWindowsFontEngine::initFontInfo(const QFontDef &request,
 
     Will probably be superseded by a common Free Type font engine in Qt 5.X.
 */
-
-QWindowsMultiFontEngine::QWindowsMultiFontEngine(QFontEngine *first, const QStringList &fallbacks)
-        : QFontEngineMulti(fallbacks.size()+1),
-          fallbacks(fallbacks)
+QWindowsMultiFontEngine::QWindowsMultiFontEngine(QFontEngine *fe, int script)
+    : QFontEngineMultiQPA(fe, script)
 {
-    if (QWindowsContext::verboseFonts)
-        qDebug() <<  __FUNCTION__ << engines.size() << first << first->fontDef.family << fallbacks;
-    engines[0] = first;
-    first->ref.ref();
-    fontDef = engines[0]->fontDef;
-    cache_cost = first->cache_cost;
-}
-
-QWindowsMultiFontEngine::~QWindowsMultiFontEngine()
-{
-    if (QWindowsContext::verboseFonts)
-        qDebug("%s", __FUNCTION__);
 }
 
 void QWindowsMultiFontEngine::loadEngine(int at)
 {
+    ensureFallbackFamiliesQueried();
     Q_ASSERT(at < engines.size());
     Q_ASSERT(engines.at(at) == 0);
 
@@ -1378,7 +1329,7 @@ void QWindowsMultiFontEngine::loadEngine(int at)
         data = fe->fontEngineData();
     }
 
-    const QString fam = fallbacks.at(at-1);
+    const QString fam = fallbackFamilyAt(at - 1);
     memcpy(lf.lfFaceName, fam.utf16(), sizeof(wchar_t) * qMin(fam.length() + 1, 32));  // 32 = Windows hard-coded
 
 #ifndef QT_NO_DIRECTWRITE
@@ -1402,9 +1353,7 @@ void QWindowsMultiFontEngine::loadEngine(int at)
                 fedw->ref.ref();
                 engines[at] = fedw;
 
-                if (QWindowsContext::verboseFonts)
-                    qDebug("%s %d %s", __FUNCTION__, at, qPrintable(fam));
-
+                qCDebug(lcQpaFonts) << __FUNCTION__ << at << fam;
                 return;
             } else {
                 qErrnoWarning("%s: CreateFontFace failed", __FUNCTION__);
@@ -1426,9 +1375,7 @@ void QWindowsMultiFontEngine::loadEngine(int at)
     engines[at] = new QWindowsFontEngine(fam, hfont, stockFont, lf, data);
     engines[at]->ref.ref();
     engines[at]->fontDef = fontDef;
-    if (QWindowsContext::verboseFonts)
-        qDebug("%s %d %s", __FUNCTION__, at, qPrintable(fam));
-
+    qCDebug(lcQpaFonts) << __FUNCTION__ << at << fam;
 
     // TODO: increase cost in QFontCache for the font engine loaded here
 }
