@@ -40,6 +40,7 @@
 ****************************************************************************/
 
 #include "qquickmenu_p.h"
+#include "qquickmenubar_p.h"
 #include "qquickmenuitemcontainer_p.h"
 #include "qquickmenupopupwindow_p.h"
 
@@ -62,7 +63,7 @@ QT_BEGIN_NAMESPACE
   \qmltype MenuPrivate
   \instantiates QQuickMenu
   \internal
-  \inqmlmodule QtQuick.Controls 1.0
+  \inqmlmodule QtQuick.Controls
  */
 
 /*!
@@ -110,7 +111,7 @@ QT_BEGIN_NAMESPACE
     \qmlproperty bool Menu::visible
 
     Whether the menu should be visible. This is only enabled when the menu is used as
-    a submenu. Its value defaults to \c true.
+    a submenu or in the menubar. Its value defaults to \c true.
 */
 
 /*!
@@ -235,7 +236,7 @@ QT_BEGIN_NAMESPACE
 */
 
 QQuickMenu::QQuickMenu(QObject *parent)
-    : QQuickMenuText(parent),
+    : QQuickMenuText(parent, QQuickMenuItemType::Menu),
       m_itemsCount(0),
       m_selectedIndex(-1),
       m_parentWindow(0),
@@ -267,10 +268,38 @@ QQuickMenu::~QQuickMenu()
             item->setParentMenu(0);
     }
 
+    if (platformItem())
+        platformItem()->setMenu(0);
+
     delete m_platformMenu;
     m_platformMenu = 0;
 }
 
+void QQuickMenu::syncParentMenuBar()
+{
+    QQuickMenuBar *menubar = qobject_cast<QQuickMenuBar *>(parent());
+    if (menubar && menubar->platformMenuBar())
+        menubar->platformMenuBar()->syncMenu(m_platformMenu);
+}
+
+void QQuickMenu::setVisible(bool v)
+{
+    QQuickMenuBase::setVisible(v);
+    if (m_platformMenu) {
+        m_platformMenu->setVisible(v);
+        syncParentMenuBar();
+    }
+}
+
+void QQuickMenu::setEnabled(bool e)
+{
+    QQuickMenuText::setEnabled(e);
+    if (m_platformMenu) {
+        m_platformMenu->setEnabled(e);
+        syncParentMenuBar();
+    }
+}
+
 void QQuickMenu::updateText()
 {
     if (m_platformMenu)
@@ -286,6 +315,8 @@ void QQuickMenu::setMinimumWidth(int w)
     m_minimumWidth = w;
     if (m_platformMenu)
         m_platformMenu->setMinimumWidth(w);
+
+    emit minimumWidthChanged();
 }
 
 void QQuickMenu::setFont(const QFont &arg)
@@ -368,8 +399,11 @@ void QQuickMenu::__popup(qreal x, qreal y, int atItemIndex)
 
     if (m_platformMenu) {
         QPointF screenPosition(x + m_xOffset, y + m_yOffset);
-        if (visualItem())
+        if (visualItem()) {
+            if (qGuiApp->isRightToLeft())
+                screenPosition.rx() -= qMax(static_cast<qreal>(m_minimumWidth), m_menuContentItem->width());
             screenPosition = visualItem()->mapToScene(screenPosition);
+        }
         m_platformMenu->showPopup(parentWindow, screenPosition.toPoint(), atItem ? atItem->platformItem() : 0);
     } else {
         m_popupWindow = new QQuickMenuPopupWindow();
@@ -377,10 +411,11 @@ void QQuickMenu::__popup(qreal x, qreal y, int atItemIndex)
             m_popupWindow->setParentItem(visualItem());
         else
             m_popupWindow->setParentWindow(parentWindow);
-        m_popupWindow->setMenuContentItem(m_menuContentItem);
+        m_popupWindow->setPopupContentItem(m_menuContentItem);
         m_popupWindow->setItemAt(atItem ? atItem->visualItem() : 0);
 
         connect(m_popupWindow, SIGNAL(visibleChanged(bool)), this, SLOT(windowVisibleChanged(bool)));
+        connect(m_popupWindow, SIGNAL(geometryChanged()), this, SIGNAL(__popupGeometryChanged()));
 
         m_popupWindow->setPosition(x + m_xOffset, y + m_yOffset);
         m_popupWindow->show();
@@ -389,8 +424,10 @@ void QQuickMenu::__popup(qreal x, qreal y, int atItemIndex)
 
 void QQuickMenu::setMenuContentItem(QQuickItem *item)
 {
-    if (m_menuContentItem != item)
+    if (m_menuContentItem != item) {
         m_menuContentItem = item;
+        emit menuContentItemChanged();
+    }
 }
 
 void QQuickMenu::setPopupVisible(bool v)
@@ -401,6 +438,14 @@ void QQuickMenu::setPopupVisible(bool v)
     }
 }
 
+QRect QQuickMenu::popupGeometry() const
+{
+    if (!m_popupWindow || !m_popupVisible)
+        return QRect();
+
+    return m_popupWindow->geometry();
+}
+
 void QQuickMenu::__closeMenu()
 {
     setPopupVisible(false);
@@ -416,7 +461,7 @@ void QQuickMenu::__dismissMenu()
     while (topMenuWindow) {
         QQuickMenuPopupWindow *pw = qobject_cast<QQuickMenuPopupWindow *>(topMenuWindow->transientParent());
         if (!pw)
-            topMenuWindow->dismissMenu();
+            topMenuWindow->dismissPopup();
         topMenuWindow = pw;
     }
 }
